# coding: utf-8
"""create or manipulate already existing labeling jobs"""

# docs
from __future__ import annotations

import time
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Union,
)

from tqdm import tqdm

if TYPE_CHECKING:
    from pandas.core.frame import DataFrame

import requests

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.api.entity_annotation.figure_api import FigureInfo
from supervisely.api.image_api import ImageInfo
from supervisely.api.module_api import (
    ApiField,
    ModuleApi,
    ModuleWithStatus,
    RemoveableBulkModuleApi,
    WaitingTimeExceeded,
)
from supervisely.collection.str_enum import StrEnum
from supervisely.geometry.alpha_mask import AlphaMask
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.point import Point
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_type import ProjectType
from supervisely.sly_logger import logger


class LabelingJobInfo(NamedTuple):
    id: int
    name: str
    readme: str
    description: str
    team_id: int
    workspace_id: int
    workspace_name: str
    project_id: int
    project_name: str
    dataset_id: int
    dataset_name: str
    created_by_id: int
    created_by_login: str
    assigned_to_id: int
    assigned_to_login: str
    reviewer_id: int
    reviewer_login: str
    created_at: str
    started_at: str
    finished_at: str
    status: str
    disabled: bool
    labeling_queue_id: int
    labeling_exam_id: int
    images_count: int
    finished_images_count: int
    rejected_images_count: int
    accepted_images_count: int
    progress_images_count: int
    classes_to_label: list
    tags_to_label: list
    images_range: tuple
    objects_limit_per_image: int
    tags_limit_per_image: int
    filter_images_by_tags: list
    include_images_with_tags: list
    exclude_images_with_tags: list
    entities: list
    priority: int


class LabelingJobApi(RemoveableBulkModuleApi, ModuleWithStatus):
    """
    API for working with Labeling Jobs. :class:`LabelingJobApi<LabelingJobApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        jobs = api.labeling_job.get_list(9) # api usage example
    """

    class Status(StrEnum):
        """Labeling Job status."""

        PENDING = "pending"
        """"""
        IN_PROGRESS = "in_progress"
        """"""
        ON_REVIEW = "on_review"
        """"""
        COMPLETED = "completed"
        """"""
        STOPPED = "stopped"
        """"""
        REVIEW_COMPLETED = "review_completed"
        """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple LabelingJobInfo information about Labeling Job.

        :Example:

         .. code-block:: python

             LabelingJobInfo(id=2,
                             name='Annotation Job (#1) (#1) (dataset_01)',
                             readme='',
                             description='',
                             team_id=4,
                             workspace_id=8,
                             workspace_name='First Workspace',
                             project_id=58,
                             project_name='tutorial_project',
                             dataset_id=54,
                             dataset_name='dataset_01',
                             created_by_id=4,
                             created_by_login='anna',
                             assigned_to_id=4,
                             assigned_to_login='anna',
                             reviewer_id=4,
                             reviewer_login='anna',
                             created_at='2020-04-08T15:10:12.618Z',
                             started_at='2020-04-08T15:10:19.833Z',
                             finished_at='2020-04-08T15:13:39.788Z',
                             status='completed',
                             disabled=False,
                             labeling_queue_id=3,
                             labeling_exam_id=None,
                             images_count=3,
                             finished_images_count=0,
                             rejected_images_count=1,
                             accepted_images_count=2,
                             progress_images_count=2,
                             classes_to_label=[],
                             tags_to_label=[],
                             images_range=(1, 5),
                             objects_limit_per_image=None,
                             tags_limit_per_image=None,
                             filter_images_by_tags=[],
                             include_images_with_tags=[],
                             exclude_images_with_tags=[],
                             entities=None,
                             priority=2)
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.README,
            ApiField.DESCRIPTION,
            ApiField.TEAM_ID,
            ApiField.WORKSPACE_ID,
            ApiField.WORKSPACE_NAME,
            ApiField.PROJECT_ID,
            ApiField.PROJECT_NAME,
            ApiField.DATASET_ID,
            ApiField.DATASET_NAME,
            ApiField.CREATED_BY_ID,
            ApiField.CREATED_BY_LOGIN,
            ApiField.ASSIGNED_TO_ID,
            ApiField.ASSIGNED_TO_LOGIN,
            ApiField.REVIEWER_ID,
            ApiField.REVIEWER_LOGIN,
            ApiField.CREATED_AT,
            ApiField.STARTED_AT,
            ApiField.FINISHED_AT,
            ApiField.STATUS,
            ApiField.DISABLED,
            ApiField.LABELING_QUEUE_ID,
            ApiField.LABELING_EXAM_ID,
            ApiField.IMAGES_COUNT,
            ApiField.FINISHED_IMAGES_COUNT,
            ApiField.REJECTED_IMAGES_COUNT,
            ApiField.ACCEPTED_IMAGES_COUNT,
            ApiField.PROGRESS_IMAGES_COUNT,
            ApiField.CLASSES_TO_LABEL,
            ApiField.TAGS_TO_LABEL,
            ApiField.IMAGES_RANGE,
            ApiField.OBJECTS_LIMIT_PER_IMAGE,
            ApiField.TAGS_LIMIT_PER_IMAGE,
            ApiField.FILTER_IMAGES_BY_TAGS,
            ApiField.INCLUDE_IMAGES_WITH_TAGS,
            ApiField.EXCLUDE_IMAGES_WITH_TAGS,
            ApiField.ENTITIES,
            ApiField.PRIORITY,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **LabelingJobInfo**.
        """
        return "LabelingJobInfo"

    def __init__(self, api):
        ModuleApi.__init__(self, api)

    def _convert_json_info(self, info: Dict, skip_missing: Optional[bool] = True):
        """ """
        if info is None:
            return None
        else:
            field_values = []
            for field_name in self.info_sequence():
                if field_name in [
                    ApiField.INCLUDE_IMAGES_WITH_TAGS,
                    ApiField.EXCLUDE_IMAGES_WITH_TAGS,
                ]:
                    continue
                value = None
                if type(field_name) is str:
                    if skip_missing is True:
                        value = info.get(field_name, None)
                    else:
                        value = info[field_name]
                elif type(field_name) is tuple:
                    for sub_name in field_name[0]:
                        if value is None:
                            if skip_missing is True:
                                value = info.get(sub_name, None)
                            else:
                                value = info[sub_name]
                        else:
                            value = value[sub_name]
                else:
                    raise RuntimeError("Can not parse field {!r}".format(field_name))

                if field_name == ApiField.FILTER_IMAGES_BY_TAGS:
                    field_values.append(value)
                    include_images_with_tags = []
                    exclude_images_with_tags = []
                    for fv in value:
                        key = ApiField.NAME
                        if key not in fv:
                            key = "title"
                        if fv["positive"] is True:
                            include_images_with_tags.append(fv[key])
                        else:
                            exclude_images_with_tags.append(fv[key])
                    field_values.append(include_images_with_tags)
                    field_values.append(exclude_images_with_tags)
                    continue
                elif (
                    field_name == ApiField.CLASSES_TO_LABEL or field_name == ApiField.TAGS_TO_LABEL
                ):
                    new_value = []
                    for fv in value:
                        key = ApiField.NAME
                        if ApiField.NAME not in fv:
                            key = "title"
                        new_value.append(fv[key])
                    field_values.append(new_value)
                    continue
                elif field_name == ApiField.IMAGES_RANGE:
                    value = (value["start"], value["end"])

                field_values.append(value)

            res = self.InfoType(*field_values)
            return LabelingJobInfo(**res._asdict())

    def _remove_batch_api_method_name(self):
        """Api remove method name."""

        return "jobs.bulk.remove"

    def _remove_batch_field_name(self):
        """Returns onstant string that represents API field name."""

        return ApiField.IDS

    def _check_membership(self, ids: List[int], team_id: int) -> None:
        """Check if user is a member of the team in which the Labeling Job is created."""
        for user_id in ids:
            memberships = self._api.user.get_teams(user_id)
            team_ids = [team.id for team in memberships]
            if team_id not in team_ids:
                raise RuntimeError(
                    f"User with id {user_id} is not a member of the team with id {team_id}"
                )

    def create(
        self,
        name: str,
        dataset_id: int,
        user_ids: List[int],
        readme: Optional[str] = None,
        description: Optional[str] = None,
        classes_to_label: Optional[List[str]] = None,
        objects_limit_per_image: Optional[int] = None,
        tags_to_label: Optional[List[str]] = None,
        tags_limit_per_image: Optional[int] = None,
        include_images_with_tags: Optional[List[str]] = None,
        exclude_images_with_tags: Optional[List[str]] = None,
        images_range: Optional[List[int, int]] = None,
        reviewer_id: Optional[int] = None,
        images_ids: Optional[List[int]] = [],
        dynamic_classes: Optional[bool] = False,
        dynamic_tags: Optional[bool] = False,
        disable_confirm: Optional[bool] = None,
        disable_submit: Optional[bool] = None,
        toolbox_settings: Optional[Dict] = None,
    ) -> List[LabelingJobInfo]:
        """
        Creates Labeling Job and assigns given Users to it.

        :param name: Labeling Job name in Supervisely.
        :type name: str
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param user_ids: User IDs in Supervisely to assign Users as labelers to Labeling Job.
        :type user_ids: List[int]
        :param readme: Additional information about Labeling Job.
        :type readme: str, optional
        :param description: Description of Labeling Job.
        :type description: str, optional
        :param classes_to_label: List of classes to label in Dataset.
        :type classes_to_label: List[str], optional
        :param objects_limit_per_image: Limit the number of objects that the labeler can create on each image.
        :type objects_limit_per_image: int, optional
        :param tags_to_label: List of tags to label in Dataset.
        :type tags_to_label: List[str], optional
        :param tags_limit_per_image: Limit the number of tags that the labeler can create on each image.
        :type tags_limit_per_image: int, optional
        :param include_images_with_tags: Include images with given tags for processing by labeler.
        :type include_images_with_tags: List[str], optional
        :param exclude_images_with_tags: Exclude images with given tags for processing by labeler.
        :type exclude_images_with_tags: List[str], optional
        :param images_range: Limit number of images to be labeled for each labeler.
        :type images_range: List[int, int], optional
        :param reviewer_id: User ID in Supervisely to assign User as Reviewer to Labeling Job.
        :type reviewer_id: int, optional
        :param images_ids: List of images ids to label in dataset
        :type images_ids: List[int], optional
        :param dynamic_classes: If True, classes created after creating the job will be available for annotators
        :type dynamic_classes: bool, optional
        :param dynamic_tags: If True, tags created after creating the job will be available for annotators
        :type dynamic_tags: bool, optional
        :param disable_confirm: If True, the Confirm button will be disabled in the labeling tool. It will remain disabled until the next API call sets the parameter to False, re-enabling the button.
        :type disable_confirm: bool, optional
        :param disable_submit: If True, the Submit button will be disabled in the labeling tool. It will remain disabled until the next API call sets the parameter to False, re-enabling the button.
        :type disable_submit: bool, optional
        :param toolbox_settings: Settings for the labeling tool. Only video projects are supported.
        :type toolbox_settings: Dict, optional
        :return: List of information about new Labeling Job. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[LabelingJobInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            user_name = 'alex'
            dataset_id = 602
            new_labeling_jobs = api.labeling_job.create(
                user_name,
                dataset_id,
                user_ids=[111, 222],
                readme='Readmy text',
                description='Work for labelers',
                objects_limit_per_image=5,
                tags_limit_per_image=3
            )
            print(new_labeling_jobs)

            # >>> List[LabelingJobInfo(id=2,...)]

            # Create video labeling job with toolbox settings

            user_id = 4
            dataset_id = 277
            video_id = 24897
            toolbox_settings = {"playbackRate": 32, "skipFramesSize": 15, "showVideoTime": True}

            new_labeling_jobs = api.labeling_job.create(
                name="Labeling Job name",
                dataset_id=dataset_id,
                user_ids=[user_id],
                readme="Labeling Job readme",
                description="Some description",
                classes_to_label=["car", "animal"],
                tags_to_label=["animal_age_group"],
                images_ids=[video_id],
                toolbox_settings=toolbox_settings,
            )
            print(new_labeling_jobs)

            # >>> List[LabelingJobInfo(id=3,...)]
        """
        if classes_to_label is None:
            classes_to_label = []
        if tags_to_label is None:
            tags_to_label = []

        filter_images_by_tags = []
        if include_images_with_tags is not None:
            for tag_name in include_images_with_tags:
                filter_images_by_tags.append({"name": tag_name, "positive": True})

        if exclude_images_with_tags is not None:
            for tag_name in exclude_images_with_tags:
                filter_images_by_tags.append({"name": tag_name, "positive": False})

        if objects_limit_per_image is None:
            objects_limit_per_image = 0

        if tags_limit_per_image is None:
            tags_limit_per_image = 0

        meta = {
            "classes": classes_to_label,
            "projectTags": tags_to_label,
            "imageTags": filter_images_by_tags,
            "imageFiguresLimit": objects_limit_per_image,
            "imageTagsLimit": tags_limit_per_image,
            "entityIds": images_ids,
            "dynamicClasses": dynamic_classes,
            "dynamicTags": dynamic_tags,
        }

        if toolbox_settings is not None:
            dataset_info = self._api.dataset.get_info_by_id(dataset_id)
            project_id = dataset_info.project_id
            project_info = self._api.project.get_info_by_id(project_id)
            project_type = project_info.type
            if project_type == ProjectType.VIDEOS.value:
                playback_rate_possible_values = [
                    0.1,
                    0.3,
                    0.5,
                    0.6,
                    0.7,
                    0.8,
                    0.9,
                    1,
                    1.1,
                    1.2,
                    1.3,
                    1.5,
                    2,
                    4,
                    8,
                    16,
                    32,
                ]
                playback_rate = toolbox_settings.get("playbackRate", None)
                if playback_rate is not None:
                    if playback_rate not in playback_rate_possible_values:
                        raise ValueError(
                            f"'playbackRate' must be one of: '{','.join(playback_rate_possible_values)}'"
                        )
                meta["toolboxSettings"] = toolbox_settings

        if disable_confirm is not None:
            meta.update({"disableConfirm": disable_confirm})
        if disable_submit is not None:
            meta.update({"disableSubmit": disable_submit})

        data = {
            ApiField.NAME: name,
            ApiField.DATASET_ID: dataset_id,
            ApiField.USER_IDS: user_ids,
            # ApiField.DESCRIPTION: description,
            ApiField.META: meta,
        }

        if readme is not None:
            data[ApiField.README] = str(readme)

        if description is not None:
            data[ApiField.DESCRIPTION] = str(description)

        if images_range is not None and images_range != (None, None):
            if len(images_range) != 2:
                raise RuntimeError("images_range has to contain 2 elements (start, end)")
            images_range = {"start": images_range[0], "end": images_range[1]}
            data[ApiField.META]["range"] = images_range

        if reviewer_id is not None:
            data[ApiField.REVIEWER_ID] = reviewer_id

        response = self._api.post("jobs.add", data)
        # created_jobs_json = response.json()

        created_jobs = []
        for job in response.json():
            created_jobs.append(self.get_info_by_id(job[ApiField.ID]))
        return created_jobs

    def get_list(
        self,
        team_id: int,
        created_by_id: Optional[int] = None,
        assigned_to_id: Optional[int] = None,
        project_id: Optional[int] = None,
        dataset_id: Optional[int] = None,
        show_disabled: Optional[bool] = False,
        reviewer_id: Optional[int] = None,
        is_part_of_queue: Optional[bool] = True,
        queue_ids: Optional[Union[List, int]] = None,
        exclude_statuses: Optional[
            List[Literal["pending", "in_progress", "on_review", "completed"]]
        ] = None,
    ) -> List[LabelingJobInfo]:
        """
        Get list of information about Labeling Job in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param created_by_id: ID of the User who created the LabelingJob.
        :type created_by_id: int, optional
        :param assigned_to_id: ID of the assigned User.
        :type assigned_to_id: int, optional
        :param project_id: Project ID in Supervisely.
        :type project_id: int, optional
        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int, optional
        :param show_disabled: Show disabled Labeling Jobs.
        :type show_disabled: bool, optional
        :param reviewer_id: ID of the User who reviews the LabelingJob.
        :type reviewer_id: int, optional
        :param is_part_of_queue: Filter by Labeling Queue. If True, all existing Labeling Jobs are returned. If False, only Labeling Jobs that are not part of the queue are returned.
        :type is_part_of_queue: bool, optional
        :param queue_ids: IDs of the Labeling Queues. If set, only Labeling Jobs from the selected queues are returned. Arg `is_part_of_queue` must be True.
        :type queue_ids: Union[List, int], optional
        :param exclude_statuses: Exclude Labeling Jobs with given statuses.
        :type exclude_statuses: List[Literal["pending", "in_progress", "on_review", "completed"]], optional
        :return: List of information about Labeling Jobs. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[LabelingJobInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            label_jobs = api.labeling_job.get_list(4)
            print(label_jobs)
            # Output: [
            #     [
            #         2,
            #         "Annotation Job (#1) (#1) (dataset_01)",
            #         "",
            #         "",
            #         4,
            #         8,
            #         "First Workspace",
            #         58,
            #         "tutorial_project",
            #         54,
            #         "dataset_01",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         "2020-04-08T15:10:12.618Z",
            #         "2020-04-08T15:10:19.833Z",
            #         "2020-04-08T15:13:39.788Z",
            #         "completed",
            #         false,
            #         3,
            #         null,
            #         3,
            #         0,
            #         1,
            #         2,
            #         2,
            #         [],
            #         [],
            #         [
            #             1,
            #             5
            #         ],
            #         null,
            #         null,
            #         [],
            #         [],
            #         [],
            #         null
            #     ],
            #     [
            #         3,
            #         "Annotation Job (#1) (#2) (dataset_02)",
            #         "",
            #         "",
            #         4,
            #         8,
            #         "First Workspace",
            #         58,
            #         "tutorial_project",
            #         55,
            #         "dataset_02",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         4,
            #         "anna",
            #         "2020-04-08T15:10:12.618Z",
            #         "2020-04-08T15:15:46.749Z",
            #         "2020-04-08T15:17:33.572Z",
            #         "completed",
            #         false,
            #         3,
            #         null,
            #         2,
            #         0,
            #         0,
            #         2,
            #         2,
            #         [],
            #         [],
            #         [
            #             1,
            #             5
            #         ],
            #         null,
            #         null,
            #         [],
            #         [],
            #         [],
            #         null
            #     ]
            # ]
        """
        if not is_part_of_queue and queue_ids is not None:
            raise ValueError("To filter by `queue_id`, `is_part_of_queue` must be set to `True`.")

        if isinstance(queue_ids, int):
            queue_ids = [queue_ids]

        filters = []
        if created_by_id is not None:
            filters.append(
                {"field": ApiField.CREATED_BY_ID[0][0], "operator": "=", "value": created_by_id}
            )
        if assigned_to_id is not None:
            filters.append(
                {"field": ApiField.ASSIGNED_TO_ID[0][0], "operator": "=", "value": assigned_to_id}
            )
        if reviewer_id is not None:
            filters.append({"field": ApiField.REVIEWER_ID, "operator": "=", "value": reviewer_id})
        if project_id is not None:
            filters.append({"field": ApiField.PROJECT_ID, "operator": "=", "value": project_id})
        if dataset_id is not None:
            filters.append({"field": ApiField.DATASET_ID, "operator": "=", "value": dataset_id})
        if not is_part_of_queue:
            filters.append({"field": ApiField.LABELING_QUEUE_ID, "operator": "=", "value": None})
        if queue_ids is not None:
            filters.append(
                {"field": ApiField.LABELING_QUEUE_ID, "operator": "in", "value": queue_ids}
            )
        if exclude_statuses is not None:
            filters.append({"field": ApiField.STATUS, "operator": "!in", "value": exclude_statuses})
        return self.get_list_all_pages(
            "jobs.list",
            {ApiField.TEAM_ID: team_id, "showDisabled": show_disabled, ApiField.FILTER: filters},
        )

    def stop(self, id: int) -> None:
        """
        Makes Labeling Job unavailable for labeler with given User ID.

        :param id: User ID in Supervisely.
        :type id: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.stop(9)
        """
        self._api.post("jobs.stop", {ApiField.ID: id})

    def get_info_by_id(self, id: int) -> LabelingJobInfo:
        """
        Get Labeling Job information by ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: Information about Labeling Job. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`LabelingJobInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            label_job_info = api.labeling_job.get_info_by_id(2)
            print(label_job_info)
            # Output: [
            #     2,
            #     "Annotation Job (#1) (#1) (dataset_01)",
            #     "",
            #     "",
            #     4,
            #     8,
            #     "First Workspace",
            #     58,
            #     "tutorial_project",
            #     54,
            #     "dataset_01",
            #     4,
            #     "anna",
            #     4,
            #     "anna",
            #     4,
            #     "anna",
            #     "2020-04-08T15:10:12.618Z",
            #     "2020-04-08T15:10:19.833Z",
            #     "2020-04-08T15:13:39.788Z",
            #     "completed",
            #     false,
            #     3,
            #     0,
            #     1,
            #     2,
            #     2,
            #     [],
            #     [],
            #     [
            #         1,
            #         5
            #     ],
            #     null,
            #     null,
            #     [],
            #     [],
            #     [],
            #     [
            #         {
            #             "reviewStatus": "rejected",
            #             "id": 283,
            #             "name": "image_03"
            #         },
            #         {
            #             "reviewStatus": "accepted",
            #             "id": 282,
            #             "name": "image_02"
            #         },
            #         {
            #             "reviewStatus": "accepted",
            #             "id": 281,
            #             "name": "image_01"
            #         }
            #     ]
            # ]
        """
        return self._get_info_by_id(id, "jobs.info")

    def archive(self, id: int) -> None:
        """
        Archives Labeling Job with given ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.archive(23)
        """
        self._api.post("jobs.archive", {ApiField.ID: id})

    def get_status(self, id: int) -> LabelingJobApi.Status:
        """
        Get status of Labeling Job with given ID.

        :param id: Labeling job ID in Supervisely.
        :type id: int
        :return: Labeling Job Status
        :rtype: :class:`Status<supervisely.api.labeling_job_api.LabelingJobApi.Status>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            job_status = api.labeling_job.get_status(4)
            print(job_status) # pending
        """
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        """ """
        # there is no ERROR status for labeling job
        pass

    def wait(
        self,
        id: int,
        target_status: str,
        wait_attempts: Optional[int] = None,
        wait_attempt_timeout_sec: Optional[int] = None,
    ) -> None:
        """
        Wait for a Labeling Job to change to the expected target status.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param target_status: Expected result status of Labeling Job.
        :type target_status: str
        :param wait_attempts: Number of attempts to retry, when :class:`WaitingTimeExceeded` raises.
        :type wait_attempts: int, optional
        :param wait_attempt_timeout_sec: Time between attempts.
        :type wait_attempt_timeout_sec: int, optional
        :raises: :class:`WaitingTimeExceeded`, if waiting time exceeded
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.wait(4, 'completed', wait_attempts=2, wait_attempt_timeout_sec=1)
            # supervisely.api.module_api.WaitingTimeExceeded: Waiting time exceeded
        """
        wait_attempts = wait_attempts or self.MAX_WAIT_ATTEMPTS
        effective_wait_timeout = wait_attempt_timeout_sec or self.WAIT_ATTEMPT_TIMEOUT_SEC
        for attempt in range(wait_attempts):
            status = self.get_status(id)
            self.raise_for_status(status)
            if status is target_status:
                return
            time.sleep(effective_wait_timeout)
        raise WaitingTimeExceeded("Waiting time exceeded")

    def get_stats(self, id: int) -> Dict:
        """
        Get stats of given Labeling Job ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: Dict with information about given Labeling Job
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            status = api.labeling_job.get_stats(3)
            print(status)
            # Output: {
            #     "job": {
            #         "editingDuration": 0,
            #         "annotationDuration": 720,
            #         "id": 3,
            #         "name": "Annotation Job (#1) (#2) (dataset_02)",
            #         "startedAt": "2020-04-08T15:15:46.749Z",
            #         "finishedAt": "2020-04-08T15:17:33.572Z",
            #         "imagesCount": 2,
            #         "finishedImagesCount": 2,
            #         "tagsStats": [
            #             {
            #                 "id": 24,
            #                 "color": "#ED68A1",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "car_color"
            #             },
            #             {
            #                 "id": 19,
            #                 "color": "#A0A08C",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "cars_number"
            #             },
            #             {
            #                 "id": 20,
            #                 "color": "#D98F7E",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "like"
            #             },
            #             {
            #                 "id": 23,
            #                 "color": "#65D37C",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "person_gender"
            #             },
            #             {
            #                 "parentId": 23,
            #                 "color": "#65D37C",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "person_gender (male)"
            #             },
            #             {
            #                 "parentId": 23,
            #                 "color": "#65D37C",
            #                 "images": 0,
            #                 "figures": 0,
            #                 "name": "person_gender (female)"
            #             },
            #             {
            #                 "id": 21,
            #                 "color": "#855D79",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "situated"
            #             },
            #             {
            #                 "parentId": 21,
            #                 "color": "#855D79",
            #                 "images": 1,
            #                 "figures": 1,
            #                 "name": "situated (inside)"
            #             },
            #             {
            #                 "parentId": 21,
            #                 "color": "#855D79",
            #                 "images": 0,
            #                 "figures": 0,
            #                 "name": "situated (outside)"
            #             },
            #             {
            #                 "id": 22,
            #                 "color": "#A2B4FA",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "vehicle_age"
            #             },
            #             {
            #                 "parentId": 22,
            #                 "color": "#A2B4FA",
            #                 "images": 0,
            #                 "figures": 1,
            #                 "name": "vehicle_age (modern)"
            #             },
            #             {
            #                 "parentId": 22,
            #                 "color": "#A2B4FA",
            #                 "images": 0,
            #                 "figures": 0,
            #                 "name": "vehicle_age (vintage)"
            #             }
            #         ]
            #     },
            #     "classes": [
            #         {
            #             "id": 43,
            #             "color": "#F6FF00",
            #             "shape": "rectangle",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "bike",
            #             "labelsCount": 0
            #         },
            #         {
            #             "id": 42,
            #             "color": "#BE55CE",
            #             "shape": "polygon",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "car",
            #             "labelsCount": 0
            #         },
            #         {
            #             "id": 41,
            #             "color": "#FD0000",
            #             "shape": "polygon",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "dog",
            #             "labelsCount": 0
            #         },
            #         {
            #             "id": 40,
            #             "color": "#00FF12",
            #             "shape": "bitmap",
            #             "totalDuration": 0,
            #             "imagesCount": 0,
            #             "avgDuration": null,
            #             "name": "person",
            #             "labelsCount": 0
            #         }
            #     ],
            #     "images": {
            #         "total": 2,
            #         "images": [
            #             {
            #                 "id": 285,
            #                 "reviewStatus": "accepted",
            #                 "annotationDuration": 0,
            #                 "totalDuration": 0,
            #                 "name": "image_01",
            #                 "labelsCount": 0
            #             },
            #             {
            #                 "id": 284,
            #                 "reviewStatus": "accepted",
            #                 "annotationDuration": 0,
            #                 "totalDuration": 0,
            #                 "name": "image_02",
            #                 "labelsCount": 0
            #             }
            #         ]
            #     }
            # }
        """
        response = self._api.post("jobs.stats", {ApiField.ID: id})
        return response.json()

    def get_activity(
        self, team_id: int, job_id: int, progress_cb: Optional[Union[tqdm, Callable]] = None
    ) -> DataFrame:
        import pandas as pd

        """
        Get all activity for given Labeling Job by ID.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param job_id: Labeling Job ID in Supervisely.
        :type job_id: int
        :param progress_cb: Function for tracking progress
        :type progress_cb: tqdm, optional
        :return: Activity data as `pd.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
        :rtype: :class:`pd.DataFrame`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            activity = api.labeling_job.get_activity(3)
            print(activity)
            # Output:
            #   userId         action  ... tagId                 meta
            # 0       4  update_figure  ...   NaN                   {}
            # 1       4  create_figure  ...   NaN                   {}
            # 2       4     attach_tag  ...  20.0                   {}
            # 3       4     attach_tag  ...  21.0  {'value': 'inside'}
            # 4       4     attach_tag  ...  24.0      {'value': '12'}
            # 5       4  update_figure  ...   NaN                   {}
            # 6       4  update_figure  ...   NaN                   {}
            # 7       4  update_figure  ...   NaN                   {}
            # 8       4  create_figure  ...   NaN                   {}
            # 9       4  update_figure  ...   NaN                   {}
            # [10 rows x 18 columns]
        """
        activity = self._api.team.get_activity(
            team_id, filter_job_id=job_id, progress_cb=progress_cb
        )
        df = pd.DataFrame(activity)
        return df

    def set_status(self, id: int, status: str) -> None:
        """
        Sets Labeling Job status.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param status: New Labeling Job status
        :type status: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.api.labeling_job_api.LabelingJobApi.Status import COMPLETED

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.labeling_job.set_status(id=9, status="completed")
        """
        self._api.post("jobs.set-status", {ApiField.ID: id, ApiField.STATUS: status})

    def get_project_meta(self, id: int) -> ProjectMeta:
        """
        Returns project meta with classes and tags used in the labeling job with given id.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :return: Project meta of the labeling job with given id.
        :rtype: :class:`ProjectMeta`
        """
        job_info = self.get_info_by_id(id)
        project_meta_json = self._api.project.get_meta(job_info.project_id)
        project_meta = ProjectMeta.from_json(project_meta_json)

        job_classes = [
            obj_class
            for obj_class in project_meta.obj_classes
            if obj_class.name in job_info.classes_to_label
        ]
        job_tags = [
            tag_meta
            for tag_meta in project_meta.tag_metas
            if tag_meta.name in job_info.tags_to_label
        ]
        job_meta = ProjectMeta(obj_classes=job_classes, tag_metas=job_tags)
        return job_meta

    def get_annotations(
        self,
        id: int,
        image_ids: Optional[List[int]] = None,
        project_meta: Optional[ProjectMeta] = None,
        image_infos: Optional[List[ImageInfo]] = None,
    ) -> List[Annotation]:
        """
        Return annotations for given image ids from labeling job with given id.
        To speed up the process, you can provide image infos, which will be used instead of fetching them from the API.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param image_ids: Image IDs in Supervisely.
                        If not provided, you must provide :param:`image_infos`.
                        Have lower priority than :param:`image_infos`.
        :type image_ids: List[int], optional
        :param project_meta: Project meta of the labeling job with given id. Can be retrieved with :func:`get_project_meta`.
        :type project_meta: :class:`ProjectMeta`, optional
        :param image_infos: List of ImageInfo objects.
                            If not provided, will be retrieved from the API.
                            Have higher priority than :param:`image_ids`.
        :type image_infos: List[ImageInfo], optional
        :return: Annotation for given image id from labeling job with given id.
        :rtype: :class:`Annotation`
        """

        def _get_geometry(type: str, data: dict):
            try:
                geometry_type = GET_GEOMETRY_FROM_STR(type)
                geometry = geometry_type.from_json(data)
            except Exception as e:
                logger.error(f"Can't parse geometry: {repr(e)}")
                geometry = None
            return geometry

        def _create_tags_from_labeling_job(
            job_tags_map: List[dict], project_meta: ProjectMeta
        ) -> List[Tag]:
            tags = []
            for tag in job_tags_map:
                tag_meta = project_meta.get_tag_meta_by_id(tag["tagId"])
                if tag_meta is None:
                    continue
                tag_value = tag.get("value", None)
                tag = Tag(tag_meta, value=tag_value)
                tags.append(tag)
            return tags

        def _create_labels_from_labeling_job(
            figures_map: List[FigureInfo], project_meta: ProjectMeta
        ) -> List[Label]:
            labels = []
            for figure in figures_map:
                obj_class = project_meta.get_obj_class_by_id(figure.class_id)
                if obj_class is None:
                    continue
                geometry = _get_geometry(figure.geometry_type, figure.geometry)
                if geometry is None:
                    continue
                tags = _create_tags_from_labeling_job(figure.tags, project_meta)
                label = Label(obj_class=obj_class, geometry=geometry, tags=tags)
                labels.append(label)
            return labels

        if image_ids is None and image_infos is None:
            raise ValueError("Either 'image_ids' or 'image_infos' must be provided.")
        if image_infos is not None:
            image_ids = [image_info.id for image_info in image_infos]
        if self._api.headers.get("x-job-id") != str(id):
            self._api.add_header("x-job-id", str(id))
        job_info = self.get_info_by_id(id)
        if image_infos is None:
            image_infos = self._api.image.get_list(
                job_info.dataset_id,
                filters=[{ApiField.FIELD: ApiField.ID, "operator": "in", "value": image_ids}],
            )
        figures_map = self._api.image.figure.download(job_info.dataset_id, image_ids)
        if self._api.headers.get("x-job-id") == str(id):
            self._api.pop_header("x-job-id")

        if project_meta is None:
            project_meta = self.get_project_meta(id)

        anns = []
        for image in image_infos:
            img_figures = figures_map.get(image.id, [])
            img_tags = _create_tags_from_labeling_job(image.tags, project_meta)
            labels = _create_labels_from_labeling_job(img_figures, project_meta)
            ann = Annotation(img_size=(image.height, image.width), labels=labels, img_tags=img_tags)
            anns.append(ann)
        return anns

    def reject_annotations(self, id: int, mode: Literal["all", "unmarked"] = "all") -> None:
        """
        Reject annotations for all or unmarked entities in the labeling job with given id.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param mode: Reject mode. Can be "all" or "unmarked".
        :type mode: str, optional
        :return: None
        :rtype: :class:`NoneType`
        """

        data = {ApiField.ID: id, ApiField.MODE: mode}
        self._api.post("jobs.reject", data)

    def set_entity_review_status(
        self, id: int, entity_id: int, status: Literal["none", "done", "accepted", "rejected"]
    ) -> None:
        """
        Sets review status for entity with given ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param entity_id: Entity ID
        :type entity_id: int
        :param status: New review status for entity
        :type status: str
        :return: None
        :rtype: :class:`NoneType`
        """
        self._api.post(
            "jobs.entities.update-review-status",
            {ApiField.JOB_ID: id, ApiField.ENTITY_ID: entity_id, ApiField.STATUS: status},
        )

    def clone(
        self,
        id: int,
        new_title: Optional[str] = None,
        reviewer_id: Optional[int] = None,
        assignee_ids: Optional[List[int]] = None,
    ) -> List[LabelingJobInfo]:
        """
        Clone Labeling Job with given ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param new_title: New title for the job
        :type new_title: str, optional
        :param reviewer_id: ID of the reviewer
        :type reviewer_id: int, optional
        :param assignee_ids: List of User IDs to assign the job
        :type assignee_ids: List[int], optional
        :return: List of information about Labeling Jobs. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[LabelingJobInfo]`
        """
        job_info = self.get_info_by_id(id)

        if new_title is None:
            new_title = job_info.name
        if reviewer_id is None:
            reviewer_id = job_info.reviewer_id
        else:
            self._check_membership([reviewer_id], job_info.team_id)
        if assignee_ids is None:
            assignee_ids = [job_info.assigned_to_id]
        else:
            self._check_membership(assignee_ids, job_info.team_id)

        job_infos = self.create(
            name=new_title,
            dataset_id=job_info.dataset_id,
            user_ids=assignee_ids,
            readme=job_info.readme,
            description=job_info.description,
            classes_to_label=job_info.classes_to_label,
            objects_limit_per_image=job_info.objects_limit_per_image,
            tags_to_label=job_info.tags_to_label,
            tags_limit_per_image=job_info.tags_limit_per_image,
            include_images_with_tags=job_info.include_images_with_tags,
            exclude_images_with_tags=job_info.exclude_images_with_tags,
            images_range=job_info.images_range,
            reviewer_id=reviewer_id,
            images_ids=[entity["id"] for entity in job_info.entities],
            # TODO dynamic_classes=job_info.dynamic_classes,
            # TODO dynamic_tags=job_info.dynamic_tags,
        )

        return job_infos

    def restart(
        self,
        id: int,
        assignee_ids: Optional[List[int]] = None,
        reviewer_id: Optional[int] = None,
        title: Optional[str] = None,
        complete_existing: Optional[bool] = True,
        only_rejected_entities: Optional[bool] = True,
        ignore_no_rejected_error: Optional[bool] = False,
    ) -> List[dict]:
        """
        Restart Labeling Job with given ID.

        :param id: Labeling Job ID in Supervisely.
        :type id: int
        :param assignee_ids: List of User IDs to assign the job. If not set, the job will be assigned to the same user as the existing job.
        :type assignee_ids: List[int], optional
        :param reviewer_id: ID of the reviewer
        :type reviewer_id: int, optional
        :param title: New title for the job <= 255 characters
        :type title: str, optional
        :param complete_existing: If False, existing job will not be completed.
        :type complete_existing: bool, optional
        :param only_rejected_entities: If False, all entities that do not have an "accepted" status will be included in new job, all unmarked entities will be rejected for the existing job.
        :type only_rejected_entities: bool, optional
        :param ignore_errors: If True, the job will not be restarted if there are errors in request data.
        :type ignore_errors: bool, optional
        :return: List of dicts with information about created Labeling Jobs.
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            api = sly.Api("https://app.supervisely.com", "your_api_token")

            job_info_list = api.labeling_job.restart(222)

            print(job_info_list)
            # Output:
            #   [
            #       {
            #           'id': 940,
            #           'userId': 342,
            #           'type': 'annotation',
            #           'name': 'Annotation Job (#2)'
            #       }
            #   ]
        """

        job_info = self.get_info_by_id(id)

        data = {ApiField.ID: id, ApiField.COMPLETE_EXISTING: complete_existing}

        if assignee_ids is not None:
            self._check_membership(assignee_ids, job_info.team_id)
            data[ApiField.USER_IDS] = assignee_ids
        else:
            data[ApiField.USER_IDS] = [job_info.assigned_to_id]

        if reviewer_id is not None:
            self._check_membership([reviewer_id], job_info.team_id)
            data[ApiField.REVIEWER_ID] = reviewer_id
        else:
            data[ApiField.REVIEWER_ID] = job_info.reviewer_id

        if title is not None:
            data[ApiField.NAME] = title

        if only_rejected_entities is False:
            self.reject_annotations(id, mode="unmarked")

        if only_rejected_entities is True and ignore_no_rejected_error is True:
            try:
                response = self._api.post("jobs.restart", data).json()
                return response
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 400 and "No images found" in e.response.text:
                    return []
                else:
                    raise e

        response = self._api.post("jobs.restart", data).json()
        return response



import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from typing import Callable, Dict, List, Union

import supervisely.io.env as sly_env
from supervisely import env, logger
from supervisely._utils import abs_url, is_development
from supervisely.api.api import Api
from supervisely.api.file_api import FileApi, FileInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.app.content import DataJson, StateJson
from supervisely.app.widgets import (
    Button,
    Checkbox,
    Container,
    Field,
    FileThumbnail,
    Flexbox,
    Input,
    ProjectThumbnail,
    Select,
    Text,
    Widget,
)
from supervisely.io.fs import get_file_name_with_ext
from supervisely.nn.artifacts.artifacts import TrainInfo

WEIGHTS_DIR = "weights"

COL_ID = "task id".upper()
COL_PROJECT = "training data".upper()
COL_CHECKPOINTS = "checkpoints".upper()
COL_SESSION = "session".upper()

columns = [
    COL_ID,
    COL_PROJECT,
    COL_CHECKPOINTS,
    COL_SESSION,
]


class CustomModelsSelector(Widget):
    class Routes:
        TASK_TYPE_CHANGED = "task_type_changed"
        VALUE_CHANGED = "value_changed"

    class ModelRow:

        def __init__(
            self,
            api: Api,
            team_id: int,
            train_info: TrainInfo,
            task_type: str,
        ):
            self._api = api
            self._team_id = team_id
            self._task_type = task_type

            task_id = train_info.task_id
            if type(task_id) is str:
                if task_id.isdigit():
                    task_id = int(task_id)
                else:
                    raise ValueError(f"Task id {task_id} is not a number")

            # col 1 task
            self._task_id = task_id
            self._task_path = train_info.artifacts_folder
            try:
                self._task_info = self._api.task.get_info_by_id(task_id)
            except:
                self._task_info = None

            if self._task_info is not None:
                self._task_date_iso = self._task_info["startedAt"]
                self._task_date = self._normalize_date()
                self._task_link = self._create_task_link()
            else:
                self._task_date_iso = None
                self._task_date = None
                self._task_link = None
            self._config_path = train_info.config_path

            # col 2 project
            self._training_project_name = train_info.project_name

            workspace_id = (
                self._task_info["workspaceId"]
                if self._task_info
                else sly_env.workspace_id(raise_not_found=False)
            )
            self._training_project_info = (
                self._api.project.get_info_by_name(workspace_id, self._training_project_name)
                if workspace_id
                else None
            )

            # col 3 checkpoints
            self._checkpoints = train_info.checkpoints

            self._checkpoints_names = []
            self._checkpoints_paths = []
            for checkpoint_info in self._checkpoints:
                if isinstance(checkpoint_info, dict):
                    self._checkpoints_names.append(checkpoint_info["name"])
                    self._checkpoints_paths.append(checkpoint_info["path"])
                elif isinstance(checkpoint_info, FileInfo):
                    self._checkpoints_names.append(checkpoint_info.name)
                    self._checkpoints_paths.append(checkpoint_info.path)

            # col 4 session
            self._session_link = train_info.session_link

            # widgets
            self._task_widget = self._create_task_widget()
            self._training_project_widget = self._create_training_project_widget()
            self._checkpoints_widget = self._create_checkpoints_widget()
            self._session_widget = self._create_session_widget()

        @property
        def task_id(self) -> int:
            return self._task_id

        @property
        def task_date(self) -> str:
            return self._task_date

        @property
        def task_link(self) -> str:
            return self._task_link

        @property
        def task_type(self) -> str:
            return self._task_type

        @property
        def training_project_info(self) -> ProjectInfo:
            return self._training_project_info

        @property
        def checkpoints_names(self) -> List[str]:
            return self._checkpoints_names

        @property
        def checkpoints_paths(self) -> List[str]:
            return self._checkpoints_paths

        @property
        def checkpoints_selector(self) -> Select:
            return self._checkpoints_widget

        @property
        def session_link(self) -> str:
            return self._session_link

        @property
        def config_path(self) -> str:
            return self._config_path

        def get_selected_checkpoint_path(self) -> str:
            return self._checkpoints_widget.get_value()

        def get_selected_checkpoint_name(self) -> str:
            return self._checkpoints_widget.get_label()

        def to_html(self) -> List[str]:
            return [
                f"<div> {self._task_widget.to_html()} </div>",
                f"<div> {self._training_project_widget.to_html()} </div>",
                f"<div> {self._checkpoints_widget.to_html()} </div>",
                f"<div> {self._session_widget.to_html()} </div>",
            ]

        def _normalize_date(self) -> str:
            date_obj = datetime.fromisoformat(self._task_date_iso.rstrip("Z"))
            formatted_date = date_obj.strftime("%d %B %Y, %H:%M")
            return formatted_date

        def _create_task_link(self) -> str:
            remote_path = os.path.join(self._task_path, "open_app.lnk")
            task_file = self._api.file.get_info_by_path(self._team_id, remote_path)
            if task_file is not None:
                if is_development():
                    return abs_url(f"/files/{task_file.id}")
                else:
                    return f"/files/{task_file.id}"
            else:
                return ""

        def _create_task_widget(self) -> Flexbox:
            if self._task_info is not None:
                task_widget = Container(
                    [
                        Text(
                            f"<i class='zmdi zmdi-folder' style='color: #7f858e'></i> <a href='{self._task_link}'>{self._task_id}</a>",
                            "text",
                        ),
                        Text(
                            f"<span class='field-description text-muted' style='color: #7f858e'>{self._task_date}</span>",
                            "text",
                            font_size=13,
                        ),
                    ],
                    gap=0,
                )
            else:
                task_widget = Text(
                    f"<span class='field-description text-muted' style='color: #7f858e'>Task was archived (ID: '{self._task_id}')</span>",
                    "text",
                )
            return task_widget

        def _create_training_project_widget(self) -> Union[ProjectThumbnail, Text]:
            if self._training_project_info is not None:
                training_project_widget = ProjectThumbnail(
                    self._training_project_info, remove_margins=True
                )
            else:
                training_project_widget = Text(
                    f"<span class='field-description text-muted' style='color: #7f858e'>Project was archived</span>",
                    "text",
                    font_size=13,
                )
            return training_project_widget

        def _create_checkpoints_widget(self) -> Select:
            checkpoint_selector_items = []
            for checkpoint_info in self._checkpoints:
                if isinstance(checkpoint_info, dict):
                    checkpoint_selector_items.append(
                        Select.Item(value=checkpoint_info["path"], label=checkpoint_info["name"])
                    )
                elif isinstance(checkpoint_info, FileInfo):
                    checkpoint_selector_items.append(
                        Select.Item(value=checkpoint_info.path, label=checkpoint_info.name)
                    )

            checkpoint_selector = Select(items=checkpoint_selector_items)
            return checkpoint_selector

        def _create_session_widget(self) -> Text:
            session_link_widget = Text(
                f"<a href='{self._session_link}'>Preview</a> <i class='zmdi zmdi-open-in-new'></i>",
                "text",
            )
            return session_link_widget

    def __init__(
        self,
        team_id: int,
        train_infos: List[TrainInfo],
        show_custom_checkpoint_path: bool = False,
        custom_checkpoint_task_types: List[str] = [],
        widget_id: str = None,
    ):
        self._api = Api.from_env()

        self._team_id = team_id

        with ThreadPoolExecutor() as executor:
            future = executor.submit(self._generate_table_rows, train_infos)
            table_rows = future.result()

        self._show_custom_checkpoint_path = show_custom_checkpoint_path
        self._custom_checkpoint_task_types = custom_checkpoint_task_types

        self._columns = columns
        self._rows = table_rows
        # self._rows_html = #[row.to_html() for row in self._rows]

        task_types = [task_type for task_type in table_rows]
        self._rows_html = defaultdict(list)
        for task_type in table_rows:
            self._rows_html[task_type].extend(
                [model_row.to_html() for model_row in table_rows[task_type]]
            )

        self._task_types = self._filter_task_types(task_types)
        if len(self._task_types) == 0:
            self.__default_selected_task_type = None
        else:
            self.__default_selected_task_type = self._task_types[0]

        self._changes_handled = False
        self._task_type_changes_handled = False

        if self._show_custom_checkpoint_path:
            self.file_thumbnail = FileThumbnail()
            team_files_url = f"{env.server_address().rstrip('/')}/files/"

            team_files_link_btn = Button(
                text="Open Team Files",
                button_type="info",
                plain=True,
                icon="zmdi zmdi-folder",
                link=team_files_url,
            )

            file_api = FileApi(self._api)
            self._model_path_input = Input(placeholder="Path to model file in Team Files")

            @self._model_path_input.value_changed
            def change_folder(value):
                file_info = None
                if value != "":
                    file_info = file_api.get_info_by_path(env.team_id(), value)
                self.file_thumbnail.set(file_info)

            model_path_field = Field(
                self._model_path_input,
                title=f"Copy path to model file from Team Files and paste to field below.",
                description="Copy path in Team Files",
            )
            self.custom_checkpoint_task_type_selector_field = None
            if len(self._custom_checkpoint_task_types) > 0:
                self.custom_checkpoint_task_type_selector_items = [
                    Select.Item(value=task_type, label=task_type)
                    for task_type in self._custom_checkpoint_task_types
                ]
                self.custom_checkpoint_task_type_selector = Select(
                    self.custom_checkpoint_task_type_selector_items
                )
                self.custom_checkpoint_task_type_selector_field = Field(
                    title="Task Type", content=self.custom_checkpoint_task_type_selector
                )

            self.custom_tab_widgets = Container(
                [
                    team_files_link_btn,
                    self.custom_checkpoint_task_type_selector_field,
                    model_path_field,
                    self.file_thumbnail,
                ]
            )

            self.custom_tab_widgets.hide()

            self.show_custom_checkpoint_path_checkbox = Checkbox("Use custom checkpoint", False)

            @self.show_custom_checkpoint_path_checkbox.value_changed
            def show_custom_checkpoint_path_checkbox_changed(is_checked):
                if is_checked:
                    self.disable_table()
                    self.custom_tab_widgets.show()
                else:
                    self.enable_table()
                    self.custom_tab_widgets.hide()

                StateJson()[self.widget_id]["useCustomPath"] = is_checked
                StateJson().send_changes()

        super().__init__(widget_id=widget_id, file_path=__file__)

    @property
    def columns(self) -> List[str]:
        return self._columns

    @property
    def rows(self) -> Dict[str, List[ModelRow]]:
        return self._rows

    def get_json_data(self) -> Dict:
        return {
            "columns": self._columns,
            "rowsHtml": self._rows_html,
            "taskTypes": self._task_types,
        }

    def get_json_state(self) -> Dict:
        return {
            "selectedRow": 0,
            "selectedTaskType": self.__default_selected_task_type,
            "useCustomPath": False,
        }

    def set_active_task_type(self, task_type: str):
        if task_type not in self._task_types:
            raise ValueError(f'Task Type "{task_type}" does not exist')
        StateJson()[self.widget_id]["selectedTaskType"] = task_type
        StateJson().send_changes()

    def get_available_task_types(self) -> List[str]:
        return self._task_types

    def disable_table(self) -> None:
        for task_type in self._rows:
            for row in self._rows[task_type]:
                row.checkpoints_selector.disable()
        super().disable()

    def enable_table(self) -> None:
        for task_type in self._rows:
            for row in self._rows[task_type]:
                row.checkpoints_selector.enable()
        super().enable()

    def enable(self):
        self.custom_tab_widgets.enable()
        self._model_path_input.enable()
        self.custom_checkpoint_task_type_selector.enable()
        self.show_custom_checkpoint_path_checkbox.enable()
        self.enable_table()
        super().enable()

    def disable(self) -> None:
        self.custom_tab_widgets.disable()
        self._model_path_input.disable()
        self.custom_checkpoint_task_type_selector.disable()
        self.show_custom_checkpoint_path_checkbox.disable()
        self.disable_table()
        super().disable()

    def _generate_table_rows(self, train_infos: List[TrainInfo]) -> Dict[str, List[ModelRow]]:
        """Method to generate table rows from remote path to training app save directory"""

        def process_train_info(train_info):
            try:
                model_row = CustomModelsSelector.ModelRow(
                    api=self._api,
                    team_id=self._team_id,
                    train_info=train_info,
                    task_type=train_info.task_type,
                )
                return train_info.task_type, model_row
            except Exception as e:
                logger.debug(f"Failed to process train info: {train_info}. Error: {repr(e)}")
                return None, None

        table_rows = defaultdict(list)

        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(process_train_info, train_info): train_info
                for train_info in train_infos
            }

            for future in as_completed(futures):
                result = future.result()
                if result:
                    task_type, model_row = result
                    if task_type is not None and model_row is not None:
                        table_rows[task_type].append(model_row)

        self._sort_table_rows(table_rows)
        return table_rows

    def _sort_table_rows(self, table_rows: Dict[str, List[ModelRow]]) -> None:
        for task_type in table_rows:
            table_rows[task_type].sort(key=lambda row: row.task_id, reverse=True)

    def _filter_task_types(self, task_types: List[str]):
        sorted_tt = []
        if "object detection" in task_types:
            sorted_tt.append("object detection")
        if "instance segmentation" in task_types:
            sorted_tt.append("instance segmentation")
        if "pose estimation" in task_types:
            sorted_tt.append("pose estimation")
        other_tasks = sorted(
            set(task_types) - set(["object detection", "instance segmentation", "pose estimation"])
        )
        sorted_tt.extend(other_tasks)
        return sorted_tt

    def get_selected_row(self, state=StateJson()) -> Union[ModelRow, None]:
        if len(self._rows) == 0:
            return
        widget_actual_state = state[self.widget_id]
        widget_actual_data = DataJson()[self.widget_id]
        task_type = widget_actual_state["selectedTaskType"]
        if widget_actual_state is not None and widget_actual_data is not None:
            selected_row_index = int(widget_actual_state["selectedRow"])
            return self._rows[task_type][selected_row_index]

    def get_selected_row_index(self, state=StateJson()) -> Union[int, None]:
        widget_actual_state = state[self.widget_id]
        widget_actual_data = DataJson()[self.widget_id]
        if widget_actual_state is not None and widget_actual_data is not None:
            return widget_actual_state["selectedRow"]

    def get_selected_task_type(self) -> str:
        return StateJson()[self.widget_id]["selectedTaskType"]

    def get_selected_model_params(self) -> Union[Dict, None]:
        config_path = None
        is_custom_path = self.use_custom_checkpoint_path()
        if not is_custom_path:
            selected_model = self.get_selected_row()
            task_type = selected_model.task_type
            checkpoint_filename = selected_model.get_selected_checkpoint_name()
            checkpoint_url = selected_model.get_selected_checkpoint_path()
            if selected_model.config_path is not None:
                config_path = selected_model.config_path
        else:
            task_type = self.get_custom_checkpoint_task_type()
            checkpoint_filename = self.get_custom_checkpoint_name()
            checkpoint_url = self.get_custom_checkpoint_path()

        model_params = {
            "model_source": "Custom models",
            "task_type": task_type,
            "checkpoint_name": checkpoint_filename,
            "checkpoint_url": checkpoint_url,
        }

        # if model_name is not None:
        #     model_params["model_name"] = model_name

        if config_path is not None:
            model_params["config_url"] = config_path

        return model_params

    # def get_selected_model_params_v2(self) -> Union[Dict, None]:

    def set_active_row(self, row_index: int) -> None:
        if row_index < 0 or row_index > len(self._rows) - 1:
            raise ValueError(f'Row with index "{row_index}" does not exist')
        StateJson()[self.widget_id]["selectedRow"] = row_index
        StateJson().send_changes()

    def use_custom_checkpoint_path(self) -> bool:
        return self.show_custom_checkpoint_path_checkbox.is_checked()

    def get_custom_checkpoint_name(self) -> str:
        if self.use_custom_checkpoint_path():
            return get_file_name_with_ext(self.get_custom_checkpoint_path())

    def get_custom_checkpoint_path(self) -> str:
        if self.use_custom_checkpoint_path():
            return self._model_path_input.get_value()

    def set_custom_checkpoint_path(self, path: str) -> None:
        if self.use_custom_checkpoint_path():
            self._model_path_input.set_value(path)

    def set_custom_checkpoint_preview(self, file_info: FileInfo) -> None:
        if self.use_custom_checkpoint_path():
            self.file_thumbnail.set(file_info)

    def get_custom_checkpoint_task_type(self) -> str:
        if self.use_custom_checkpoint_path():
            return self.custom_checkpoint_task_type_selector.get_value()

    def set_custom_checkpoint_task_type(self, task_type: str) -> None:
        if self.use_custom_checkpoint_path():
            available_task_types = self.custom_checkpoint_task_type_selector.get_labels()
            if task_type not in available_task_types:
                raise ValueError(f'"{task_type}" is not available task type')
            self.custom_checkpoint_task_type_selector.set_value(task_type)

    def task_type_changed(self, func: Callable):
        route_path = self.get_route_path(CustomModelsSelector.Routes.TASK_TYPE_CHANGED)
        server = self._sly_app.get_server()
        self._task_type_changes_handled = True

        @server.post(route_path)
        def _task_type_changed():
            res = self.get_selected_task_type()
            func(res)

        return _task_type_changed

    def value_changed(self, func: Callable):
        route_path = self.get_route_path(CustomModelsSelector.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_selected_row()
            func(res)

        return _value_changed



# coding: utf-8

import hashlib
import os
import os.path as osp
import shutil

from supervisely.io import fs as sly_fs


# stateless object
class FSCache:
    def __init__(self, name, storage_root):
        self._name = name
        self._storage_root = storage_root

    @classmethod
    def _copy_file_concurr(cls, src_path, dst_path):
        try:
            sly_fs.copy_file(src_path, dst_path)
        except OSError as e:
            # may be written by parallel process, skip in that case
            if not osp.isfile(dst_path):
                raise e

    @classmethod
    def _copy_dir_recursively(cls, src_path, dst_path):
        if sly_fs.dir_exists(dst_path):
            sly_fs.remove_dir(dst_path)
        shutil.copytree(src_path, dst_path)

    @classmethod
    def _get_obj_suffix(cls, st_path):
        _, suffix = osp.splitext(st_path.rstrip("/"))
        return suffix

    def _storage_obj_exists(self, st_path, suffix):
        raise NotImplementedError()

    def _get_suffix(self, path):
        raise NotImplementedError()

    def _write_obj_impl(self, src_path, st_path):
        raise NotImplementedError()

    def _read_obj_impl(self, st_path, dst_path):
        raise NotImplementedError()

    def _rm_obj_impl(self, st_path):
        raise NotImplementedError()

    @property
    def storage_root_path(self):
        return self._storage_root

    # slow
    def list_objects(self):
        def scan_deeper(paths):
            res = []
            for left_part in paths:
                for right_part in os.listdir(left_part):
                    new_p = osp.join(left_part, right_part)
                    res.append(new_p)
            return res

        obj_paths = [self._storage_root]
        for lvl in range(3):  # fixed, determined by get_storage_path impl
            obj_paths = scan_deeper(obj_paths)

        obj_pathes_suffixes = [(p, self._get_suffix(p)) for p in obj_paths]
        return obj_pathes_suffixes

    def get_storage_path(self, data_hash, suffix=""):
        if suffix:
            suffix = ".{}".format(suffix).replace("..", ".")
        st_hash = hashlib.sha256(data_hash.encode("utf-8")).hexdigest()
        st_path = osp.join(self._storage_root, st_hash[0:2], st_hash[2:5], st_hash + suffix)
        return st_path

    def check_storage_object(self, data_hash, suffix=""):
        st_path = self.get_storage_path(data_hash, suffix)
        if self._storage_obj_exists(st_path, suffix):
            return st_path
        return None

    def write_object(self, src_path, data_hash):
        suffix = self._get_suffix(src_path)
        st_path = self.get_storage_path(data_hash, suffix)
        if not self._storage_obj_exists(st_path, suffix):
            self._write_obj_impl(src_path, st_path)

    def write_objects(self, src_paths, datas_hashes, progress_cb=None):
        for src_path, data_hash in zip(src_paths, datas_hashes):
            self.write_object(src_path, data_hash)
            if progress_cb is not None:
                progress_cb()

    def read_object(self, data_hash, dst_path):
        suffix = self._get_suffix(dst_path)
        st_path = self.check_storage_object(data_hash, suffix)
        if not st_path:
            return None  # doesn't exist
        self._read_obj_impl(st_path, dst_path)
        return dst_path

    # @TODO: to remove??? used only in data_manager->download project
    def read_objects(self, dst_paths_hashes, progress_ctr):
        written_paths = []
        for dst_path, data_hash in dst_paths_hashes:
            res_path = self.read_object(data_hash, dst_path)
            if res_path:
                written_paths.append(res_path)
                progress_ctr.iter_done_report()
        return written_paths

    def remove_object(self, st_path, suffix=""):
        removable = self._storage_obj_exists(st_path, suffix)
        if removable:
            self._rm_obj_impl(st_path)
        return removable


# class ImageStorage(FSStorage):
class FileCache(FSCache):
    def _storage_obj_exists(self, st_path, suffix):
        if not suffix:
            raise ValueError("Storage {}. File ext is empty.".format(self._name))
        return osp.isfile(st_path)

    def _get_suffix(self, path):
        return sly_fs.get_file_ext(path)

    def _write_obj_impl(self, src_path, st_path):
        sly_fs.ensure_base_path(st_path)
        self._copy_file_concurr(src_path, st_path)

    def _read_obj_impl(self, st_path, dst_path):
        sly_fs.ensure_base_path(dst_path)
        sly_fs.copy_file(st_path, dst_path)

    def _rm_obj_impl(self, st_path):
        os.remove(st_path)

    async def _read_obj_impl_async(self, st_path, dst_path):
        sly_fs.ensure_base_path(dst_path)
        await sly_fs.copy_file_async(st_path, dst_path)

    async def write_object_async(self, src_path, data_hash):
        suffix = self._get_suffix(src_path)
        st_path = self.get_storage_path(data_hash, suffix)
        if not self._storage_obj_exists(st_path, suffix):
            await sly_fs.copy_file_async(src_path, st_path)

    async def read_object_async(self, data_hash, dst_path):
        suffix = self._get_suffix(dst_path)
        st_path = self.check_storage_object(data_hash, suffix)
        if not st_path:
            return None
        await self._read_obj_impl(st_path, dst_path)
        return dst_path


class NNCache(FSCache):
    def _storage_obj_exists(self, st_path, suffix):
        if suffix:
            raise ValueError("Storage {}. Unexpected suffix for NN dir.".format(self._name))
        return osp.isdir(st_path)

    def _get_suffix(self, path):
        return ""

    def _write_obj_impl(self, src_path, st_path):
        self._copy_dir_recursively(src_path, st_path)

    def _read_obj_impl(self, st_path, dst_path):
        self._copy_dir_recursively(st_path, dst_path)

    def _rm_obj_impl(self, st_path):
        sly_fs.remove_dir(st_path)


class EmptyCache(FSCache):
    def _storage_obj_exists(self, st_path, suffix):
        return False

    def _get_suffix(self, path):
        return ""

    def _write_obj_impl(self, src_path, st_path):
        raise NotImplementedError()

    def _read_obj_impl(self, st_path, dst_path):
        raise NotImplementedError()

    def _rm_obj_impl(self, st_path):
        raise NotImplementedError()

    def write_objects(self, src_paths, datas_hashes, progress_cb=None):
        raise NotImplementedError()

    def read_objects(self, dst_paths_hashes, progress_ctr):
        return []  # overridden to speed up



import asyncio
import os
import time

from tqdm import tqdm

from supervisely import Api, logger
from supervisely._utils import run_coroutine
from supervisely.io.fs import clean_dir, ensure_base_path

# LOG_LEVEL = "INFO"
LOG_LEVEL = "DEBUG"
api = Api.from_env()

api.logger.setLevel(LOG_LEVEL)

TEAM_ID = 10
user_path = os.path.expanduser("~")
source_path = f"{user_path}/test_project_download/"
save_path = f"{user_path}/Work/test_project_download/"
logger.info(f"Save path: {save_path}")
remote_path = "/videos/"
remote_upload_test_sync = "/test_sync/"
remote_upload_test_async = "/test_async/"
ensure_base_path(save_path)
clean_dir(save_path)
files = (
    ("12391768_3840_2160_30fps.mp4", f"{remote_path}12391768_3840_2160_30fps.mp4"),
    ("MOV_h264.mov", f"{remote_path}MOV_h264.mov"),
    ("MP4_HEVC.mp4", f"{remote_path}MP4_HEVC.mp4"),
)
file_name = "vids.tar"


async def download_files():
    sema = asyncio.Semaphore(10)
    tasks = []
    for name, path in files:
        progress = tqdm(total=None, desc=f"Downloading {name}", unit="B", unit_scale=True)
        task = api.file.download_async(TEAM_ID, path, save_path + name, sema, progress_cb=progress)
        tasks.append(task)
    await asyncio.gather(*tasks)


def main_db():
    remote_paths = [path for _, path in files]
    names = [name for name, _ in files]
    sizeb_list = [api.file.get_info_by_path(TEAM_ID, path).sizeb for _, path in files]
    sizeb = sum(sizeb_list)
    save_paths = [save_path + name for name in names]
    progress = tqdm(total=sizeb, desc="Downloading files", unit="B", unit_scale=True)
    download_coro = api.file.download_bulk_async(
        TEAM_ID, remote_paths, save_paths, progress_cb=progress
    )
    run_coroutine(download_coro)


def maind_df():
    run_coroutine(download_files())


def main_ip():
    start = time.time()
    # os.environ["CONTEXT_SLYFILE"] = "/test/test_input.tar"
    os.environ["FOLDER"] = "/videos/"
    os.environ["TEAM_ID"] = str(TEAM_ID)
    download_coro = api.file.download_input_async(save_path, show_progress=True)
    run_coroutine(download_coro)
    end = time.time()
    print(f"Time taken for download input async: {end-start}")


def main_dd():
    start = time.time()
    download_coro = api.file.download_directory_async(
        TEAM_ID, remote_path, save_path, semaphore=asyncio.Semaphore(100)
    )
    run_coroutine(download_coro)
    end = time.time()
    print(f"Time taken for download dir async: {end-start}")


def compare_dir_download():
    main_dd()

    clean_dir(save_path)

    start = time.time()
    api.file.download_directory(TEAM_ID, remote_path, save_path)
    end = time.time()
    print(f"Time taken for download dir: {end-start}")


def upload():
    total = os.path.getsize(source_path + file_name)
    process_cb = tqdm(total=total, desc="Uploading files", unit="B", unit_scale=True)
    upload_coro = api.file.upload_async(
        TEAM_ID,
        source_path + file_name,
        remote_path + file_name,
        progress_cb=process_cb,
    )
    start = time.monotonic()
    run_coroutine(upload_coro)
    end = time.monotonic()
    logger.info(f"Time taken for upload async: {(end-start) / 60}")


def upload_bulk():
    total = 0
    src_paths = []
    dst_paths = []
    for file in os.listdir(source_path):
        path = os.path.join(source_path, file)
        src_paths.append(path)
        dst_paths.append(remote_path + file)
        total += os.path.getsize(path)
    process_cb = tqdm(total=total, desc="Uploading files", unit="B", unit_scale=True)
    upload_coro = api.file.upload_bulk_async(
        TEAM_ID,
        src_paths,
        dst_paths,
        progress_cb=process_cb,
    )
    start = time.monotonic()
    run_coroutine(upload_coro)
    end = time.monotonic()
    logger.info(f"Time taken for upload bulk async: {(end-start) / 60}")


def main_ud():
    total = 0
    for file in os.listdir(source_path):
        path = os.path.join(source_path, file)
        total += os.path.getsize(path)

    process_cb = tqdm(total=total, desc="Uploading files", unit="B", unit_scale=True)
    start = time.monotonic()
    api.file.upload_directory(
        TEAM_ID,
        source_path,
        remote_upload_test_sync,
        progress_size_cb=process_cb,
    )
    end = time.monotonic()
    logger.info(f"Time taken for upload dir: {end-start}")

    process_cb_as = tqdm(total=total, desc="Uploading files async", unit="B", unit_scale=True)
    coro = api.file.upload_directory_async(
        TEAM_ID,
        source_path,
        remote_upload_test_async,
        progress_size_cb=process_cb_as,
    )
    start = time.monotonic()
    run_coroutine(coro)
    end = time.monotonic()
    logger.info(f"Time taken for upload dir async: {end-start}")


if __name__ == "__main__":
    try:
        # maind_df()  # to download and save files
        # main_dd()  # to download and save files as folder
        # main_db()  # to download and save files in bulk
        # main_ip()  # to download input file
        # compare_dir_download()  # to compare download time between async and sync
        main_ud()
    except KeyboardInterrupt:
        logger.info("Stopped by user")



# coding: utf-8

from collections import namedtuple

import numpy as np

from supervisely.annotation.label import Label, PixelwiseScoresLabel
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.point_location import PointLocation
from supervisely.annotation.tag import Tag


def segmentation_array_to_sly_bitmaps(idx_to_class: dict, pred: np.ndarray, origin: PointLocation=None) -> list:
    """
    Converts array with segmentation results to Labels with Bitmap geometry according to idx_to_class mapping.

    Args:
        idx_to_class: Dict matching values in prediction array with appropriate ObjClass.
        pred: Array containing raw segmentation results.
        origin: Origin point for all output Bitmaps.
    return:
        A list containing result labels.
    """
    labels = []
    for cls_idx, cls_obj in idx_to_class.items():
        predicted_class_pixels = (pred == cls_idx)
        if np.any(predicted_class_pixels):
            class_geometry = Bitmap(data=predicted_class_pixels, origin=origin)
            labels.append(Label(geometry=class_geometry, obj_class=cls_obj))
    return labels


def segmentation_scores_to_per_class_labels(idx_to_class: dict, segmentation_scores: np.ndarray) -> list:
    """
    Converts output network segmentation scores into list of PixelwiseScoresLabels with MultichannelBitmap geometry.

    Args:
        idx_to_class: Dict matching channels in prediction scores array with appropriate ObjClass.
        segmentation_scores: Array containing raw segmentation scores.
    Returns:
        A list containing PixelwiseScoresLabel(s) with MultichannelBitmap(s).
    """
    return [
        PixelwiseScoresLabel(
            geometry=MultichannelBitmap(data=segmentation_scores[:, :, cls_idx, np.newaxis]),
            obj_class=cls_obj)
        for cls_idx, cls_obj in idx_to_class.items()]


DetectionNetworkPrediction = namedtuple('DetectionNetworkPrediction', ['boxes', 'scores', 'classes'])


def detection_preds_to_sly_rects(idx_to_class, network_prediction: DetectionNetworkPrediction, img_shape,
                                 min_score_threshold, score_tag_meta) -> list:
    """
    Converts network detection results to Supervisely Labels with Rectangle geometry.

    Args:
        idx_to_class: Dict matching predicted boxes with appropriate ObjClass.
        network_prediction: Network predictions packed into DetectionNetworkPrediction instance.
        img_shape: Size(height, width) of image that was used for inference.
        min_score_threshold: All detections with less scores will be dropped.
        score_tag_meta: TagMeta instance for score tags.
    Returns:
        A list containing labels with detection rectangles.
    """
    labels = []
    thr_mask = np.squeeze(network_prediction.scores) > min_score_threshold
    for box, class_id, score in zip(np.squeeze(network_prediction.boxes)[thr_mask],
                                    np.squeeze(network_prediction.classes)[thr_mask],
                                    np.squeeze(network_prediction.scores)[thr_mask]):

            xmin = round(float(box[1] * img_shape[1]))
            ymin = round(float(box[0] * img_shape[0]))
            xmax = round(float(box[3] * img_shape[1]))
            ymax = round(float(box[2] * img_shape[0]))

            rect = Rectangle(top=ymin, left=xmin, bottom=ymax, right=xmax)
            class_obj = idx_to_class[int(class_id)]
            label = Label(geometry=rect, obj_class=class_obj)

            score_tag = Tag(score_tag_meta, value=round(float(score), 4))
            label = label.add_tag(score_tag)
            labels.append(label)
    return labels



import os
import supervisely as sly
import re

import tqdm

import traceback
from rich.console import Console
import os


def download_directory_run(
    team_id: int,
    remote_dir: str,
    local_dir: str,
    filter: str = None,
    ignore_if_not_exists: bool = False,
) -> bool:
    console = Console()

    api = sly._handle_creds_error_to_console(sly.Api.from_env, console.print)
    if api is False:
        return False

    if api.team.get_info_by_id(team_id) is None:
        console.print(
            f"\nError: Team with ID={team_id} is either doesn't exist or not found in your account\n",
            style="bold red",
        )
        return False

    # force directories to end with slash '/'
    if not local_dir.endswith("/"):
        local_dir += "/"
    if not remote_dir.endswith("/"):
        remote_dir += "/"

    files = api.file.listdir(team_id, remote_dir, recursive=True)
    if len(files) == 0:
        if ignore_if_not_exists:
            console.print(
                f"\nWarning: Team files folder '{remote_dir}' doesn't exist. Skipping command...\n",
                style="bold yellow",
            )
            return True
        console.print(
            f"\nError: Team files folder '{remote_dir}' doesn't exist\n", style="bold red"
        )
        return False

    if filter is not None:
        filtered = [f for f in files if bool(re.search(filter, sly.fs.get_file_name_with_ext(f)))]

    console.print(
        f"\nDownloading directory '{remote_dir}' from Team files ...\n",
        style="bold",
    )

    try:
        if filter is not None:
            with tqdm.tqdm(desc="Downloading from Team files...", total=len(filtered)) as pbar:
                for remote_path in filtered:
                    local_save_path = os.path.join(
                        local_dir,
                        os.path.dirname(remote_path).strip("/"),
                        os.path.basename(remote_path),
                    )
                    api.file.download(
                        team_id,
                        remote_path,
                        local_save_path,
                    )
                    pbar.update(1)
        else:
            total_size = api.file.get_directory_size(team_id, remote_dir)
            pbar = tqdm.tqdm(
                desc="Downloading from Team files...", total=total_size, unit="B", unit_scale=True
            )
            api.file.download_directory(team_id, remote_dir, local_dir, progress_cb=pbar)

        console.print(
            f"\nTeam files directory has been sucessfully downloaded to the local path: '{local_dir}'.\n",
            style="bold green",
        )
        return True

    except:
        console.print("\nError: Download failed\n", style="bold red")
        traceback.print_exc()
        return False



from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class YOLOv8(BaseTrainArtifacts):
    def __init__(self, team_id: int):
        super().__init__(team_id)

        self._app_name = "Train YOLOv8 | v9 | v10 | v11"
        self._slug = "supervisely-ecosystem/yolov8/train"
        self._serve_app_name = "Serve YOLOv8 | v9 | v10 | v11"
        self._serve_slug = "supervisely-ecosystem/yolov8/serve"
        self._framework_name = "YOLOv8"
        self._framework_folder = "/yolov8_train"
        self._weights_folder = "weights"
        self._task_type = None
        self._weights_ext = ".pt"
        self._config_file = None
        self._pattern = re_compile(
            r"^/yolov8_train/(object detection|instance segmentation|pose estimation)/[^/]+/\d+/?$"
        )
        self._available_task_types: List[str] = [
            "object detection",
            "instance segmentation",
            "pose estimation",
        ]
        self._require_runtime = True

    def get_task_id(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        return parts[-1]

    def get_project_name(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        return parts[-2]

    def get_task_type(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        return parts[2]

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)

    def get_config_path(self, artifacts_folder: str) -> str:
        return None



# coding: utf-8

import json


NONE_STR = 'None'


def str_to_type_or_none(data_str, target_type):
    if data_str == NONE_STR:
        return None
    res = json.loads(data_str)
    if type(res) != target_type:
        raise TypeError("variable has to be of type {!r} or None".format(target_type))
    return res



# coding: utf-8
"""load and inference models"""

from __future__ import annotations
import os
from os import PathLike
from typing import List, Union

import numpy as np
import requests

import supervisely.io.env as sly_env
import supervisely.io.json as sly_json
from supervisely.api.module_api import ApiField
from supervisely.api.task_api import TaskApi
from supervisely.nn.experiments import ExperimentInfo
from supervisely.nn.model.prediction import Prediction
from supervisely.nn.model.prediction_session import PredictionSession
from supervisely.nn.utils import ModelSource
from supervisely.project.project_meta import ProjectMeta
from supervisely.api.api import Api


class ModelAPI:
    def __init__(self, api: "Api" = None, task_id: int = None, url: str = None):
        assert not (task_id is None and url is None), "Either `task_id` or `url` must be passed."
        assert (
            task_id is None or url is None
        ), "Either `task_id` or `url` must be passed (not both)."
        if task_id is not None:
            assert api is not None, "API must be provided if `task_id` is passed."

        self.api = api
        self.task_id = task_id
        self.url = url

        if self.task_id is not None:
            task_info = self.api.task.get_info_by_id(self.task_id)
            if task_info is None:
                raise ValueError(f"Task with id {self.task_id} not found.")
            self.url = f'{self.api.server_address}/net/{task_info["meta"]["sessionToken"]}'

    # region Requests
    def _post(self, method: str, data: dict, raise_for_status: bool = True):
        url = f"{self.url.rstrip('/')}/{method.lstrip('/')}"
        response = requests.post(url, json=data)
        if raise_for_status:
            response.raise_for_status()
        return response.json()

    def _get(self, method: str, params: dict = None, raise_for_status: bool = True):
        url = f"{self.url.rstrip('/')}/{method.lstrip('/')}"
        response = requests.get(url, params=params)
        if raise_for_status:
            response.raise_for_status()
        return response.json()

    # ------------------------------------ #

    # region Info
    def get_info(self):
        if self.task_id is not None:
            return self.api.nn._deploy_api.get_deploy_info(self.task_id)
        return self._post("get_deploy_info", {})

    def get_settings(self):
        if self.task_id is not None:
            return self.api.task.send_request(self.task_id, "get_custom_inference_settings", {})[
                "settings"
            ]
        else:
            return self._post("get_custom_inference_settings", {})["settings"]

    def get_model_meta(self):
        if self.task_id is not None:
            return ProjectMeta.from_json(
                self.api.task.send_request(self.task_id, "get_output_classes_and_tags", {})
            )
        else:
            return ProjectMeta.from_json(self._post("get_output_classes_and_tags", {}))

    def get_classes(self):
        model_meta = self.get_model_meta()
        return [obj_class.name for obj_class in model_meta.obj_classes]

    def list_pretrained_models(self) -> List[str]:
        """Return a list of pretrained model names available for deployment"""
        return self._post("list_pretrained_models", {})

    def list_pretrained_model_infos(self) -> List[dict]:
        """Return a list of pretrained model infos with full information about each model"""
        return self._post("list_pretrained_model_infos", {})

    def list_experiments(self) -> List[ExperimentInfo]:
        """Return a list of training experiments in Supervisely"""
        raise NotImplementedError

    def is_deployed(self) -> bool:
        if self.task_id is not None:
            return self.api.task.is_ready(self.task_id)
        return self._post("is_ready", {})["status"] == "ready"

    def status(self):
        if self.task_id is not None:
            return self.api.task.send_request(self.task_id, "get_status", {})
        return self._post("get_status", {})

    def shutdown(self):
        if self.task_id is not None:
            return self.api.task.stop(self.task_id)
        response = self._post("tasks.stop", {ApiField.ID: id})
        return TaskApi.Status(response[ApiField.STATUS])

    # --------------------- #

    # region Load
    def load(
        self,
        model: str,
        device: str = None,
        runtime: str = None,
    ):
        if self.url is not None:
            if os.path.exists(model):
                self._load_local_custom_model(model, device, runtime)
            else:
                self._load_local_pretrained_model(model, device, runtime)

        elif model.startswith("/"):
            team_id = sly_env.team_id()
            artifacts_dir, checkpoint_name = (
                self.api.nn._deploy_api._get_artifacts_dir_and_checkpoint_name(model)
            )
            self.api.nn._deploy_api.load_custom_model(
                self.task_id,
                team_id,
                artifacts_dir,
                checkpoint_name,
                device=device,
                runtime=runtime,
            )
        else:
            self.api.nn._deploy_api.load_pretrained_model(
                self.task_id, model, device=device, runtime=runtime
            )

    def _load_local_pretrained_model(self, model: str, device: str = None, runtime: str = None):
        available_models = self.list_pretrained_models()
        if model not in available_models:
            raise ValueError(f"Model {model} not found in available models: {available_models}")

        deploy_params = {
            "model_files": {},
            "model_source": ModelSource.PRETRAINED,
            "model_info": {},
            "device": device,
            "runtime": runtime,
        }
        state = {"deploy_params": deploy_params, "model_name": model}
        return self._post("deploy_from_api", {"state": state})

    def _load_local_custom_model(self, model: str, device: str = None, runtime: str = None):
        deploy_params = self._get_custom_model_params(model, device, runtime)
        state = {"deploy_params": deploy_params, "model_name": model}
        return self._post("deploy_from_api", {"state": state})

    def _get_custom_model_params(self, model_name: str, device: str = None, runtime: str = None):
        def _load_experiment_info(artifacts_dir):
            experiment_path = os.path.join(artifacts_dir, "experiment_info.json")
            model_info = sly_json.load_json_file(experiment_path)
            model_meta_path = os.path.join(artifacts_dir, "model_meta.json")
            model_info["model_meta"] = sly_json.load_json_file(model_meta_path)
            original_model_files = model_info.get("model_files")
            return model_info, original_model_files

        def _prepare_local_model_files(artifacts_dir, checkpoint_path, original_model_files):
            return {k: os.path.join(artifacts_dir, v) for k, v in original_model_files.items()} | {
                "checkpoint": checkpoint_path
            }

        model_source = ModelSource.CUSTOM
        artifacts_dir = os.path.dirname(os.path.dirname(model_name))
        model_info, original_model_files = _load_experiment_info(artifacts_dir)
        model_files = _prepare_local_model_files(artifacts_dir, model_name, original_model_files)
        deploy_params = {
            "model_files": model_files,
            "model_source": model_source,
            "model_info": model_info,
            "device": device,
            "runtime": runtime,
        }
        return deploy_params

    # --------------------------------- #

    # region Predict
    def predict_detached(
        self,
        input: Union[np.ndarray, str, PathLike, list] = None,
        image_id: int = None,
        video_id: int = None,
        dataset_id: int = None,
        project_id: int = None,
        batch_size: int = None,
        conf: float = None,
        classes: List[str] = None,
        upload_mode: str = None,
        **kwargs,
    ) -> PredictionSession:
        if upload_mode is not None:
            kwargs["upload_mode"] = upload_mode
        return PredictionSession(
            self.url,
            input=input,
            image_id=image_id,
            video_id=video_id,
            dataset_id=dataset_id,
            project_id=project_id,
            api=self.api,
            batch_size=batch_size,
            conf=conf,
            classes=classes,
            **kwargs,
        )

    def predict(
        self,
        input: Union[np.ndarray, str, PathLike, list] = None,
        image_id: Union[List[int], int] = None,
        video_id: Union[List[int], int] = None,
        dataset_id: Union[List[int], int] = None,
        project_id: Union[List[int], int] = None,
        batch_size: int = None,
        conf: float = None,
        img_size: int = None,
        classes: List[str] = None,
        upload_mode: str = None,
        recursive: bool = None,
        **kwargs,
    ) -> List[Prediction]:
        if "show_progress" not in kwargs:
            kwargs["show_progress"] = True
        if recursive is not None:
            kwargs["recursive"] = recursive
        if img_size is not None:
            kwargs["img_size"] = img_size
        return list(
            self.predict_detached(
                input,
                image_id,
                video_id,
                dataset_id,
                project_id,
                batch_size,
                conf,
                classes,
                upload_mode,
                **kwargs,
            )
        )

    # ------------------------------------ #



# coding: utf-8
"""list all user roles that are available on private supervisely instance"""

# docs
from __future__ import annotations

from enum import IntEnum
from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApiBase


class RoleInfo(NamedTuple):
    id: int
    role: str
    created_at: str
    updated_at: str


class RoleApi(ModuleApiBase):
    """
    API for working with Roles. :class:`RoleApi<RoleApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervisely.com", token="4r47N...xaTatb")

        roles = api.role.get_list() # api usage example
    """

    class DefaultRole(IntEnum):
        """class DefaultRole"""

        ADMIN = 1
        """"""
        DEVELOPER = 2
        """"""
        ANNOTATOR = 3
        """"""
        VIEWER = 4
        """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple RoleInfo information about Role.

        :Example:

         .. code-block:: python

            RoleInfo(id=71,
                     role='manager',
                     created_at='2019-12-10T14:31:41.878Z',
                     updated_at='2019-12-10T14:31:41.878Z')
        """
        return [ApiField.ID, ApiField.ROLE, ApiField.CREATED_AT, ApiField.UPDATED_AT]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **RoleInfo**.
        """
        return "RoleInfo"

    def get_list(self, filters: Optional[List[Dict[str, str]]] = None) -> List[RoleInfo]:
        """
        List of all roles that are available on private Supervisely instance.

        :param filters: List of params to sort output Roles.
        :type filters: list
        :return: List of all roles with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[RoleInfo]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            roles = api.role.get_list()
        """
        return self.get_list_all_pages("roles.list", {ApiField.FILTER: filters or []})

    def _convert_json_info(self, info: dict, skip_missing=True):
        """ """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return RoleInfo(**res._asdict())



from logging import Logger
from pathlib import Path
from typing import Tuple, Union


def pages_to_images(
    doc_path: Union[Path, str],
    save_path: Union[Path, str],
    dpi: int,
    logger: Logger,
    filetype: str = "pdf",
) -> bool:
    import fitz # pylint: disable=import-error

    doc_path = Path(doc_path)
    save_path = Path(save_path)

    name, is_pdf = get_file_name_and_check_type(doc_path, filetype)

    if not is_pdf:
        logger.warning(
            f"The file extention for document {name} is wrong or there is no extension founded."
        )
        logger.warning("Trying to read as PDF.")

    try:
        doc = fitz.Document(filename=doc_path, filetype=filetype)
    except:
        logger.warn(f"Skip converting {name} file: file is broken or have wrong extension.")
        return False

    for page in doc:
        pix = page.get_pixmap(dpi=dpi)
        pix_save_path = save_path / f"{name}_page_{page.number}.png"
        pix.save(pix_save_path)
    return True


def get_file_name_and_check_type(doc_path: Path, filetype: str = "pdf") -> Tuple[str, bool]:
    suffix = doc_path.suffix
    if len(suffix) > 0:
        suffix = suffix[1:]

    if suffix == filetype:
        return doc_path.stem, True
    else:
        return doc_path.name, False
